import os
import warnings
from dataclasses import dataclass, field
from typing import Optional, Union, List, Dict

from sistem.genome import hg38_chrom_lengths_from_cytoband, get_chrom_lens_from_reference

@dataclass
class Parameters:
    """
    Global parameter values for a simulation experiment.

    Attributes:
        N0 (int): The initial number of cells at time t=0 and upon site seeding. **Default: 10.**
        growth_rate (float): The primary site exponential growth rate. **Default: 0.0051.**
        max_growth_rate_multiplier (int, float): This parameter is used to increase the exponential growth rates in metastatic sites. Metastatic growth rates are at minimum *growth_rate* but are increased based on the fitness of the cell which initiates seeding, up to a maximum of *growth_rate* * *max_growth_rate_multiplier*. **Default: 5.**
        capacities (int, list): The carrying capacity (max number of cells) of each anatomical site. If an integer is provided, all sites have that same carrying capacity. Also accepts a list of carrying capacities, one for each site. **Default: 1e7.**
        epsilon (float): A per-cell per-generation baseline migration probability to a site. **Default: 1e-8.**
        lifespan_mean (int, float): The mean cell lifespan in number of generations drawn from an exponential distribution. Corresponds to 1/r, where r is the rate parameter of the exponential. A value of 1 will fix the lifespan to 1 generation for all cells. 

        ref (str, optional): Path to an input reference genome in fasta format. Can be used to initialize chromosome sizes, but is required for generating synthetic sequencing reads. **Default: None.**
        alt_ref (str, optional): Path to an optionally alternate reference genome in fasta format. Allele 0 utilizes ref, while Allele 1 utilizes alt_ref. Use if the goal is to generate allele-specific synthetic sequencing reads. **Default: None.**
        chrom_names (list, optional): A list of chromosome names to use in the reference sequence(s). Only necessary to specify if the reference genome contains superfluous chromosomes which should not be included in the genome. **Default: None.**
        chrom_lens (dict, int, optional): A dictionary describing the size of the genome where keys are chromosome names and values are chromosome lengths in number of base pairs. If kept as None, will automatically populate with chr1-chr22 and lengths derived from the hg38 human reference genome. If ref is specified, will automatically populate based on the provided sequences. Additionally, if an integer is passed, then all chromosomes 1-*num_chrom* (see below) will have the same given length. **Default: None.**
        num_chrom (int): The number of chromosomes in the genome. Use only if an int is passed to *chrom_lens*, or if you want to use the first *num_chrom* human reference chromosomes if *chrom_lens* is set to None. Otherwise, will update accordingly. **Default: 22.**
        region_len (int): SISTEM utilizes a simplified genome representation whereby chromosome sequences are partitioned into non-overlapping regions of uniform size *region_len* (in base pairs). Higher values reduce memory burden, while lower values increase simulation resolution. **Default: 5e6.**

        arm_ratios (float, dict): The ratio of the small chromosome arm length to the total chrosome length. If chrom_lens is None, will utilize arm ratios derived from the hg38 human reference genome. Can pass a dictionary where keys are chromosome names and values are ratios, or a single ratio used by all chromosomes. **Default: 0.5.**

        focal_driver_rate (float): The probability of acquiring a driver focal (segmental) CNA at each generation. **Default: 1e-4.**
        focal_pass_rate (float): The probability of acquiring a passenger focal (segmental) CNA at each generation. **Default: = 0.01.**
        length_mean (int, float): The mean number of regions a focal CNA spans. Length is drawn from an exponential distribution. **Default = 1.5.**
        focal_gain_rate (float): The probability that a focal CNA is amplification (gain) versus a deletion (loss) **Default = 0.5.**
        mag_mean (int, float): The mean number of additional copies gained during a focal amplification CNA. Amplification magnitude is drawn from a geometric distribution. Default = 1.2.**
        SNV_driver_rate (float): The probability of acquiring a driver SNV at each generation. **Default: 1e-4**
        SNV_pass_rate (float): The probability of acquiring a passenger SNV at each generation. **Default: 0.01.**
        arm_rate (float): The probability of acquiring a chromosome-arm CNA at each generation. **Default: 1e-5.**
        chromosomal_rate (float): The probability of acquiring a whole-chromosomal CNA at each generation. **Default: 1e-6.**
        chrom_dup_rate (float): The probability that a chromosome-arm CNA or a whole-chromosomal CNA is a duplication versus a deletion. **Default: 1e-5.**
        WGD_rate (float): The probability of acquiring a WGD at each generation. **Default: 1e-8.**

        CN_coeff (float): The maximum CN selection coefficient magnitude. Used only for random initialization. **Default: 0.25.**
        SNV_coeff (float): The maximum region SNV selection coefficient magnitude. Used only for random initialization. **Default: 0.1.**
        OG_r (float): The ratio of regions which are OGs. Used only for random initialization in the Region/Hybrid Selection Model. **Default: 0.05.**
        TSG_r (float): The ratio of regions which are TSGs. Used only for random initialization in the Region/Hybrid Selection Model.. **Default: 0.05.**
        EG_r (float): The ratio of regions which are EGs (essential genes). Used only for random initialization in the Hybrid Selection Model. **Default: 0.05.**
        alter_prop (float): Parameter for creating site-specific metastatic libraries. Represents the fraction of driver selection coefficients to alter in each site if method is 'random', or of the farthest site if method is 'distance', with the number of altered coefficients scaled accordingly for the rest. **Default: 0.3.**

        max_region_CN (int): Viability checkpoint parameter. Represents the maximum CN of a driver region. **Default: 10.**
        max_region_SNV (int): Viability checkpoint parameter. Represents the maximum number of driver SNVs in a single region. **Default: 10.**
        max_ploidy (int, float): Viability checkpoint parameter. Represents the maximum ploidy of a cell. **Default: 8.**
        min_ploidy (int, float): Viability checkpoint parameter. Represents the minimum ploidy of a cell. **Default: 1.5.**
        max_distinct_driv_ratio (float): Viability checkpoint parameter. Represents the maximum number of distinct drivers which can be mutated. **Default: 0.8.**

        t_max (int): The maximum number of generations to run. **Default: 6000.**
        min_detectable (int): Terminates the growth simulator when the number of cells present in each anatomical site is atleast *min_detectable*. **Default: 5e-5.**
        ncells_prim (int): The number of cells to sample from the primary site. **Default: 100.**
        ncells_meta (int): The number of cells to sample from the metastatic sites. **Default: 100.**
        ncells_normal (int): The number of normal cells to diluate the primary site with. If generating clonal lineages, a relative number of normal cells will be sampled from the metastatic sites as well. Passing 1 will add a convenient normal cell outgroup to the simulated lineage tree. **Default: 1.**
        min_mut_fraction (float): In SISTEM, clones are defined by cells with a unique sequence of driver mutations, but this loose definition means that distinct 'clones' appearing in the clonal lineage tree may differ by a just a few small mutations. The *min_mut_fraction* parameter can help make the clones present in the tree more distinct. It describes the minimum frequency a clone's genotype must occur in the sampled cells to remain in the tree. If possible, multiple clones in a subtree will merge together with a common genotype to remain above *min_mut_fraction*, otherwise they are pruned. **Default: 0.05**.
 
        bin_size (int, optional): The size in base pairs of the copy number windows/segments in the final output profiles. Essentially groups consecutive regions together into larger bins and computes the mean. If kept as None, the bin size will be set equal to the *region_len*. This will be desirable for the majority of cases. Only specify if simulating with a region length that is smaller than practical (e.g. if *region_len* is less than 100kbp for single-cell data). **Default: None.**
        coverage (int, float): The average number of reads which cover any given base pair in the genome. When used to generate single-cell read counts or DNA-seq reads, it is recommended to use a low value (<0.2), whereas if used to generate bulk read counts, it is recommended to use a high value (>50). **Default: 0.1.**
        read_len (int): The length of the paired-end reads. Together with coverage, used to compute expected read counts. **Default: 150.**
        seq_error (float): Per-base pair sequencing error rate. Only used for generating scDNA-seq reads **Default: 0.02.**
        lorenz_y (float): Used to introduce coverage non-uniformity when generating single-cell read counts and DNA-seq. In a nutshell, coverage uniformity is parameterized by a point on the lorenz curve (0.5, *lorenz_y*). Default value of 0.5 means maximally uniform, and decreasing *lorenz_y* down to 0 will decrease uniformity. Only specify if evaluating conditions under non-uniform coverage. Be aware that this parameter is extremely sensitive, and values <=0.4 will lead to highly non-uniform distributions. **Default: 0.5.**

        out_dir (str): The path to the output directly. **Default: 'out'.**
        log_path (str, optional): The path to the log file. By default, will write to *out_dir*/sim.log. **Default: None.**
        num_processors (int): Number of processors to use when generating synthetic scDNA-seq reads. Will not speed up the other steps of the simulator. **Default: 1.**

    """
    N0: int = 10
    growth_rate: float = 0.0051
    max_growth_rate_multiplier: Union[int, float] = 5
    capacities: Union[int, List[int]] = field(default_factory=lambda: [1e7])
    nsites: int = 1
    epsilon: float = 1e-8
    lifespan_mean: Union[int, float] = 1
    
    ref: Optional[str] = None
    alt_ref: Optional[str] = None
    chrom_names: Optional[List[str]] = None
    chrom_lens: Optional[Union[int, Dict]] = None
    num_chroms: int = 22
    region_len: int = field(default=int(5e6))
    arm_ratios: Union[float, Dict] = field(default=0.5)

    focal_driver_rate: float = 5e-4
    SNV_driver_rate: float = 5e-4
    arm_rate: float = 1e-4
    chromosomal_rate: float = 5e-5
    WGD_rate: float = 1e-8
    focal_gain_rate: float = 0.5
    chrom_dup_rate: float = 0.5
    length_mean: Union[int, float] = 1.5
    mag_mean: Union[int, float] = 1.2

    focal_pass_rate: float = 0.01
    SNV_pass_rate: float = 0.01

    # region library
    CN_coeff: float = 0.25
    SNV_coeff: float = 0.1
    OG_r: float = 0.05
    TSG_r: float = 0.05
    EG_r: float = 0.05
    alter_prop: float = 0.3

    max_region_CN: int = 10
    max_region_SNV: int = 10
    max_ploidy: Union[int, float] = 8
    min_ploidy: Union[int, float] = 1.5
    max_distinct_driv_ratio: float = 0.8

    t_max: int = 6000
    min_detectable: int = 5e5
    ncells_prim: int = 100
    ncells_meta: int = 100
    ncells_normal: int = 1
    min_mut_fraction: float = 0.05

    bin_size: Optional[int] = None
    coverage: Union[int, float] = 0.1
    read_len: int = 150
    seq_error: float = 0.02
    lorenz_y: float = 0.5

    out_dir: str = 'out'
    log_path: Optional[str] = None
    num_processors: int = 1

    def __post_init__(self):
        if self.nsites < 1 or self.nsites > 100:
            raise ValueError("Number of sites must be greater than 0 and less than 100.")
        if self.nsites >= 15:
            warnings.warn("Extremely large number of sites. Consider choosing a smaller value.")

        self.region_len = int(self.region_len)
        self._process_genome()

        if self.lifespan_mean < 0:
            raise ValueError("Lifespan must be >= 1.")

        if self.CN_coeff < 0 or self.CN_coeff > 1:
            raise ValueError("CN_coeff must be between 0-1.")
        
        if self.SNV_coeff < 0 or self.SNV_coeff > 1:
            raise ValueError("SNV_coeff must be between 0-1.")

        if self.alter_prop < 0 or self.alter_prop > 1:
            raise ValueError("alter_prop must be between 0-1.")

        if self.growth_rate <= 0:
            raise ValueError("growth_rate must be greater than 0.")
        if self.max_growth_rate_multiplier < 1:
            raise ValueError("max_growth_rate_multiplier must be 1 or greater.")
        if self.bin_size is None or self.bin_size < self.region_len:
            self.bin_size = self.region_len
        else:
            self.bin_size = int(self.bin_size)
        
        if self.ncells_prim > self.min_detectable or self.ncells_meta > self.min_detectable:
            raise ValueError("min_detectable must be greater than ncells_prim and ncells_meta")
        
        if self.out_dir is None:
            raise ValueError("out_dir must be provided.")

        if self.out_dir is not None:
            if not os.path.isdir(self.out_dir):
                os.makedirs(self.out_dir)
        
        if self.log_path is None:
            self.log_path = os.path.join(self.out_dir, 'sim.log')
    
    def _process_genome(self):
        if self.chrom_names is not None:
            for chrname in self.chrom_names:
                if 'X' in chrname or 'Y' in chrname:
                    warnings.warn("Sex chromosomes detected in provided chrom names, but SISTEM treats all chromosomes as autosomes.")
                    break
            self.num_chroms = len(self.chrom_names)
        if self.num_chroms > 22:
            warnings.warn("Large number of chromosomes detected (>22).")

        if self.ref is not None:
            if not os.path.exists(self.ref):
                raise ValueError("Cannot find file at the provided ref path.")
            chrom_lens = get_chrom_lens_from_reference(self.ref, chrom_names=self.chrom_names)
            if len(chrom_lens) > 22:
                warnings.warn("Provided reference genome contains a large number of chromosomes (>22). Consider specifying which chromosomes to use from the reference genome with the chrom_names parameter. Example: chrom_names = ['chr1', 'chr2', ..., 'chr22'].")
            for chrname in chrom_lens:
                if 'X' in chrname or 'Y' in chrname:
                    warnings.warn("Provided reference genome may contain sex chromosomes (SISTEM treats all chromosomes as autosomes). Consider specifying which chromosomes to use from the reference genome with the chrom_names parameter. Example: chrom_names = ['chr1', 'chr2', ..., 'chr22'].")
            self.chrom_lens = chrom_lens
            if len(chrom_lens) == 22:
                cl, ar = hg38_chrom_lengths_from_cytoband()
                self.arm_ratios = ar

            if self.alt_ref is not None:
                if not os.path.exists(self.alt_ref):
                    raise ValueError("Cannot find file at the provided alt_ref path.")
                alt_chrom_lens = get_chrom_lens_from_reference(self.alt_ref, chrom_names=self.chrom_names)
                if set(chrom_lens.keys()) != set(alt_chrom_lens).keys():
                    raise ValueError("Chromosome names in ref and alt_ref do not match.")
        
        elif self.chrom_lens is None:
            cl, ar = hg38_chrom_lengths_from_cytoband()
            if self.num_chroms > 22:
                raise ValueError("When no reference is passed and chrom_lens is None, hg38 chrom lengths are used. Cannot set num_chroms > 22.")
            if self.num_chroms < 22:
                chrnames = [f'chr{i}' for i in range(1, self.num_chroms + 1)]
                cl = {i: v for i,v in cl.items() if i in chrnames}
                ar = {i: v for i,v in ar.items() if i in chrnames}
            self.chrom_lens = cl
            self.arm_ratios = ar
        
        elif isinstance(self.chrom_lens, int):
            if self.chrom_names is not None:
                self.chrom_lens = {c: self.chrom_lens for c in self.chrom_names}
            else:
                self.chrom_lens = {f'chr{i}': self.chrom_lens for i in range(1, self.num_chroms + 1)}

        elif isinstance(self.chrom_lens, dict):
            self.num_chroms = len(self.chrom_lens)


def fill_params(params, **kwargs):
    """
    Create a Parameters by updating the `params` object based on provided keyword arguments, or creating a new one with the keywords.

    Arguments:
        params: The Parameters object to update.
        **kwargs: Keyword arguments where the keys correspond to the parameter names 
                  to be updated and the values are the new values for those parameters.

    Returns:
        A new Parameters object with updated values.
    """
    # Create a copy of the original params to avoid modifying the original
    if params is None:
        keywords = {}
        for key, value in kwargs.items():
            if value is not None:
                keywords[key] = value
        updated_params = Parameters(**keywords)
        return updated_params
    else:
        params_dict = params.__dict__.copy()
        for key, value in kwargs.items():
            if key in params_dict and value is not None:
                params_dict[key] = value
        updated_params = Parameters(**params_dict)
        return updated_params


import copy
import numpy as np
from typing import Optional

from sistem.genome.genome import Genome
from sistem.selection import BaseLibrary, Attributes
from sistem.lineage.mutation import CNA, SNV
from sistem.lineage.tree import Node

class Cell(Node):
    gen = None

    def __init__(
        self, 
        birth_gen: Optional[int] = None, 
        genome: Optional[Genome] = None, 
        library: Optional[BaseLibrary] = None, 
        attributes: Optional[Attributes] = None, 
        site: int = 0, 
        **kwargs
    ):
        self.birth_gen = birth_gen
        self.genome = genome
        self.library = library
        self.attributes = attributes
        self.site = site
        self.events = []
        self.birth_counts = {}
        super().__init__(**kwargs)
    
    def inherit(self):
        assert self.parent is not None, 'Cannot inherit if parent is None'
        self.genome = copy.deepcopy(self.parent.genome)

    def add_SNV(self, chromosome, start, position, bp, record_event=True, driver=False):
        chromosome.add_SNV(start, position, bp)
        if record_event:
            self.events.append(SNV(chrom=chromosome.name, allele=chromosome.allele, homolog_id=chromosome.homolog_id, index=start, region=chromosome.seq[start], position=position, bp=bp, driver=driver, gen=self.gen))

    ## For passengers
    def focal_amplification(self, chromosome, start, end, m, record_event=True, driver=False):
        chromosome.amplify(start, end, m)
        if record_event:
            self.events.append(CNA(category='focal_amplification', chrom=chromosome.name, allele=chromosome.allele, homolog_id=chromosome.homolog_id, start=start, end=end, copies=m, gene_ids=chromosome.seq[start:end], driver=driver, gen=self.gen))
    
    ## For passengers
    def focal_deletion(self, chromosome, start, end, record_event=True, driver=False):
        chromosome.delete(start, end)
        if record_event:
            self.events.append(CNA(category='focal_deletion', chrom=chromosome.name, allele=chromosome.allele, homolog_id=chromosome.homolog_id, start=start, end=end, gene_ids=chromosome.seq[start:end], driver=driver, gen=self.gen))
    
    def arm_duplication(self, chromosome, arm, record_event=True):
        chrname = chromosome.name
        new_homolog_id = len(self.genome[chrname])
        nchromosome = chromosome.copy(homolog_id=new_homolog_id)
        self.genome[chrname].append(nchromosome)
        if nchromosome.cent:
            if arm == 'p':
                nchromosome.delete(nchromosome.cent, len(nchromosome))
            elif arm == 'q':
                nchromosome.delete(0, nchromosome.cent)
        if record_event:
            self.events.append(CNA(category='arm_duplication', chrom=chromosome.name, allele=chromosome.allele, homolog_id=chromosome.homolog_id, spawn_id=new_homolog_id, arm=arm, gen=self.gen))
        
    def arm_deletion(self, chromosome, arm, record_event=True):
        if chromosome.cent:
            if arm == 'p':
                self.focal_deletion(chromosome, 0, chromosome.cent, record_event=False)
            elif arm == 'q':
                self.focal_deletion(chromosome, chromosome.cent, len(chromosome), record_event=False)
        else:
            self.focal_deletion(chromosome, 0, len(chromosome), record_event=False)
        if record_event:
            self.events.append(CNA(category='arm_deletion', chrom=chromosome.name, allele=chromosome.allele, homolog_id=chromosome.homolog_id, arm=arm, gen=self.gen))

    def chromosomal_duplication(self, chromosome, record_event=True):
        chrname = chromosome.name
        new_homolog_id = len(self.genome[chrname])
        self.genome[chrname].append(chromosome.copy(homolog_id=new_homolog_id))
        if record_event:
            self.events.append(CNA(category='chromosomal_duplication', chrom=chromosome.name, allele=chromosome.allele, homolog_id=chromosome.homolog_id, spawn_id=new_homolog_id, gen=self.gen))

    def chromosomal_deletion(self, chromosome, record_event=True):
        self.focal_deletion(chromosome, 0, len(chromosome), record_event=False)
        if record_event:
            self.events.append(CNA(category='chromosomal_deletion', chrom=chromosome.name, allele=chromosome.allele, homolog_id=chromosome.homolog_id, gen=self.gen))

    def WGD(self, record_event=True):
        for chrname in self.genome.chrom_names:
            n = len(self.genome[chrname])
            new_chroms = []
            for chromosome in self.genome[chrname]:
                if len(chromosome) > 0:
                    new_chroms.append(chromosome.copy(homolog_id=n))
                    n += 1
            for nchrom in new_chroms:
                self.genome[chrname].append(nchrom)
        if record_event:
            self.events.append(CNA(category='WGD', gen=self.gen))

    def get_driver_start_regions(self, chromosome, size):
        '''
        Returns a list of all possible start locations such that an event of a given size will hit atleast one driver regions.
        '''
        return self.library.get_driver_start_regions(self, chromosome, size)
    
    def get_passenger_start_regions(self, chromosome, size):
        '''
        Returns a list of all possible start locations such that an event of a given size will not hit any driver regions.
        '''
        return self.library.get_passenger_start_regions(self, chromosome, size)

class Clone(Cell):
    def __init__(
        self, 
        popsize: int = 1, 
        fitness: Optional[float] = None, 
        **kwargs
    ):
        super().__init__(**kwargs)
        self.popsize = popsize
        self.death_gen = None
        self.fitness = fitness
        if self.library and self.attributes is None:
            self.library.init_attributes(self)

    def __len__(self):
        return self.popsize

    def set_library(self, library):
        self.library = library
        self.library.init_attributes(self)
    
    def replicate(self, site=None, popsize=1, library=None, name=None):
        if site == None:
            site = self.site
        if library == None:
            library = self.library

        if library != self.library:
            attributes = self.library.init_attributes(self)
        else:
            attributes = copy.deepcopy(self.attributes)

        new_clone = Clone(name=name, parent=self, genome=copy.deepcopy(self.genome), library=library, site=site, popsize=popsize, attributes=attributes, birth_gen=self.gen, fitness=self.fitness)

        return new_clone
    
    def die(self):
        self.popsize = 0
        self.death_gen = Clone.gen
        # Do not remove genome if object is the starting clone
        if self.birth_gen == 0:
            self.genome = None
            self.attributes = None

    def get_ploidy(self):
        return self.genome.nregions / self.library._ndiploid_regions

    def is_viable(self):
        return self.library.check_viability(self)
    
    def update_fitness(self):
        '''
        Call the library-specific fitness function model
        '''
        self.fitness = self.library.compute_fitness(self)

    def get_birth_prob(self, mean_fit, Ntot, Etot):
        return (self.fitness / mean_fit) * (Etot / (Etot + Ntot))
    
    def get_num_births(self, mean_fit, Ntot, Etot):
        birth_prob = self.get_birth_prob(mean_fit, Ntot, Etot)
        if birth_prob >= 1:
            nbirths = self.popsize
        else:
            nbirths = np.random.binomial(self.popsize, birth_prob)
        return nbirths
    
    def add_SNV(self, chromosome, start, position, bp, record_event=True):
        self.library.update_stats(self, chromosome, start, start)
        super().add_SNV(chromosome, start, position, bp, record_event=record_event, driver=True)

    def focal_amplification(self, chromosome, start, end, m, record_event=True):
        '''
        Given a chromosome, start and end indicies, and a magnitude, amplify the region and update the cell fitness to reflect changes in driver regions
        '''
        self.genome.nregions += m * (end - start)
        self.library.update_stats(self, chromosome, start, end, mag=m)
        super().focal_amplification(chromosome, start, end, m, record_event=record_event, driver=True)

    def focal_deletion(self, chromosome, start, end, record_event=True):
        self.genome.nregions -= (end - start)
        self.library.update_stats(self, chromosome, start, end, mag=-1)
        super().focal_deletion(chromosome, start, end, record_event=record_event, driver=True)

    def arm_duplication(self, chromosome, arm, record_event=True):
        if chromosome.cent:
            if arm == 'p':
                self.genome.nregions += chromosome.cent
                self.library.update_stats(self, chromosome, 0, chromosome.cent)
            elif arm == 'q':
                self.genome.nregions += (len(chromosome) - chromosome.cent)
                self.library.update_stats(self, chromosome, chromosome.cent, len(chromosome))
        else:
            self.genome.nregions += len(chromosome)
            self.library.update_stats(self, chromosome, 0, len(chromosome))
        super().arm_duplication(chromosome, arm, record_event=record_event)
        
    def chromosomal_duplication(self, chromosome, record_event=True):
        self.genome.nregions += len(chromosome)
        self.library.update_stats(self, chromosome, 0, len(chromosome))
        super().chromosomal_duplication(chromosome, record_event=record_event)

    def WGD(self, record_event=True):
        self.genome.nWGD += 1
        self.genome.nregions *= 2
        for chrname in self.genome.chrom_names:
            for chromosome in self.genome[chrname]:
                self.library.update_stats(self, chromosome, 0, len(chromosome))
        super().WGD(record_event=record_event)


from sistem.selection.base_library import BaseLibrary, Attributes
from sistem.selection.arm_library import RandomArmLibrary, FittedArmLibrary
from sistem.selection.region_library import RandomRegionLibrary, FittedRegionLibrary
from sistem.selection.hybrid_library import RandomHybridLibrary, FittedHybridLibrary

__all__ = [
    'BaseLibrary',
    'Attributes',
    'RandomArmLibrary',
    'FittedArmLibrary',
    'RandomRegionLibrary',
    'FittedRegionLibrary',
    'RandomHybridLibrary',
    'FittedHybridLibrary'
]


import numpy as np
from typing import Optional, Union, List

from sistem.anatomy.base_anatomy import BaseAnatomy
from sistem.selection import BaseLibrary

class StaticAnatomy(BaseAnatomy):
    """The static migration model.
    
    In this model, the probabiltiy of migrations from site :math:`a` to site :math:`b` depend on the distance :math:`d(a,b)`, but are the same for each cell in the site at each generation. 
    """
    def __init__(self, libraries: Optional[Union[BaseLibrary, List[BaseLibrary]]], **kargs):
        super().__init__(libraries=libraries, **kargs)
        self.tmatrix = None
    
    def initialize_distances(self, method='random', matrix=None, path=None):
        super().initialize_distances(method=method, matrix=matrix, path=path)
        
        rates = [(1/x)*self.eps if x != 0 else 0 for x in self.dists[:, 0]]
        idxs = np.triu_indices(self.nsites)
        self._tmatrix = np.zeros((self.nsites, self.nsites))
        self._tmatrix[idxs] = rates
        self._tmatrix = self._tmatrix + self._tmatrix.T - np.diag(np.diag(self._tmatrix))
        self.initialized = True

    def get_oneway_transition_probabilities(self, clone):
        return self._tmatrix[clone.site]
    
    def get_total_transition_probability(self, clone):
        sprobs = self._tmatrix[clone.site]
        tot_prob = 1 - np.prod([1 - x for x in sprobs])
        return tot_prob


import numpy as np
import copy
import os
from typing import Optional, Union, Dict

from sistem.selection.base_library import BaseLibrary, Attributes
from sistem.anatomy.utils import assign_n_alt_drivers
from sistem.utilities.utilities import get_reg_id
from sistem.parameters import Parameters, fill_params

class BaseRegionLibrary(BaseLibrary):
    """
    Base region library

    Attributes:
        drivers (dict): keys are chromosome names and values are a list of ids of OGs and TSGs
        essential (dict): keys are chromosome names and values are a list of ids of ESSs
    """
    is_driver_region_model = True
    is_driver_SNV_model = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.drivers = {chrname: [] for chrname in self.regions.keys()}
        self.essential = {chrname: [] for chrname in self.regions.keys()}
    
    # Needed
    def compute_fitness(self, clone):
        """
        Update the fitness coefficient of the genome by traversing over every key region
        """
        fitness = 1
        ploidy = clone.get_ploidy()
        for chrname in self.regions:
            for r in self.drivers[chrname]:
                s = self[chrname][r]
                n = clone.attributes.driver_counts[chrname][r]
                fitness *= (1 + s)**(n/ploidy)
        return fitness
    
    # Needed
    def init_base_fit(self):
        self.base_fit = 1
        for chrname, regions in self.drivers.items():
            for r in regions:
                s = self[chrname][r]
                self.base_fit *= (1 + s)
    
    def init_max_fit(self):
        def compute_fitness_from_counts(driv_counts, ploidy):
            if ploidy > self.max_ploidy or ploidy < self.min_ploidy:
                return -1
            fitness = 1
            for chrname in self.regions:
                for r,n in driv_counts[chrname].items():
                    s = self[chrname][r]
                    fitness *= (1 + s)**(n/ploidy)
            return fitness

        OG, TSG = [], []
        for chrname,deltas in self.delta.items():
            for r,s in enumerate(deltas):
                if s > 0:
                    OG.append((chrname,r,s))
                elif s < 0:
                    TSG.append((chrname,r,s))
        OG.sort(key=lambda x: x[2], reverse=True)
        TSG.sort(key=lambda x: x[2])

        driv_counts = {chrname: {r: 2 for r in self.drivers[chrname]} for chrname in self.regions}
        cur_fit = self.base_fit
        start_nregions = self._ndiploid_regions
        cur_nregions = start_nregions*2
        count = 0
        while count < self.max_distinct_driv and (len(OG) > 0 or len(TSG) > 0):
            fit1, fit2 = -1, -1
            if len(OG) > 0:
                chrname,r,s = OG[0]
                driv_counts[chrname][r] = self.max_region_CN
                nregions = cur_nregions + self.max_region_CN - 2
                fit1 = compute_fitness_from_counts(driv_counts, nregions/start_nregions)
                driv_counts[chrname][r] = 2
            if len(TSG) > 0:
                chrname,r,s = TSG[0]
                driv_counts[chrname][r] = 0
                nregions = cur_nregions - 2
                fit2 = compute_fitness_from_counts(driv_counts, nregions/start_nregions)
                driv_counts[chrname][r] = 2
            if fit1 == -1 and fit2 == -1:
                break
            elif fit1 > fit2:
                chrname,r,s = OG.pop(0)
                driv_counts[chrname][r] = self.max_region_CN
                cur_nregions = cur_nregions + self.max_region_CN - 2
                cur_fit = fit1
            else:
                chrname,r,s = TSG.pop(0)
                driv_counts[chrname][r] = 0
                nregions = cur_nregions - 2
                cur_fit = fit2
            count += 1
        self.max_fit = cur_fit

    def check_viability(self, clone):
        p = 2**(clone.genome.nWGD + 1)
        
        driv_counts = [c for chrname in self.regions for r,c in clone.attributes.driver_counts[chrname].items() if c != p]
        nmutated_drivers = len(driv_counts)
        if len(driv_counts) == 0:
            max_CN = p
        else:
            max_CN = max(driv_counts)
        ploidy = clone.get_ploidy()

        if nmutated_drivers > self.max_distinct_driv:
            return False
        elif ploidy > self.max_ploidy:
            return False
        elif ploidy < self.min_ploidy:
            return False
        elif max_CN > self.max_region_CN:
            return False
        else:
            return True
    
    # Needed
    def update_stats(self, clone, chromosome, start, end, mag=1):
        for r in chromosome.seq[start:end]:
            if r in self.drivers[chromosome.name] + self.essential[chromosome.name]:
                clone.attributes.driver_counts[chromosome.name][r] += mag
    
    # Needed
    def init_attributes(self, clone):
        driver_counts = {chrname: {r: 0 for r in self.drivers[chrname] + self.essential[chrname]} for chrname in self.regions}
        if clone.genome is not None:
            for chrname in self.regions:
                for chromosome in clone.genome[chrname]:
                    for r in chromosome.seq:
                        if r in self.drivers[chrname]:
                            driver_counts[chrname][r] += 1
        clone.attributes = Attributes(driver_counts=driver_counts)
    
    # Needed
    def get_driver_start_regions(self, cell, chromosome, size):
        chrname = chromosome.name
        sequence = chromosome.seq
        possible_start = set()
        for i,r in enumerate(sequence[:-size]):
            # Must be > 0 because it is still in the seq
            if r in self.drivers[chrname] + self.essential[chrname]:
                for j in range(max(0, i - size + 1), i+1):
                    possible_start.add(j)
        return list(possible_start)

    # Needed
    def get_passenger_start_regions(self, cell, chromosome, size):
        """
        Returns a list of all possible start locations such that an event of a given size will not hit any driver regions
        """
        chrname = chromosome.name
        sequence = [get_reg_id(r) for r in chromosome.seq]
        n = len(sequence) - size + 1
        
        starts = []
        i = 0

        while i < n:
            hits = [j for j,r in enumerate(sequence[i:i+size], i) if r in self.drivers[chrname] + self.essential[chrname]]
            if len(hits) > 0:
                i = max(hits) + 1
            else:
                while i < n and sequence[i+size-1] not in self.drivers[chrname] + self.essential[chrname]:
                    starts.append(i)
                    i += 1

        return starts
    
    def create_metastatic_libraries(self, nsites, alter_prop, CN_coeff, method='random', dists=None):
        if method == 'random':
            return self.create_metastatic_libraries_random(nsites, alter_prop, CN_coeff)
        elif method == 'distance':
            if dists is None:
                raise ValueError("Cannot create metastatic libraries with the 'distance' method when dists has not been initialized.")
            return self.create_metastatic_libraries_distance(nsites, alter_prop, CN_coeff, dists)
        else:
            raise ValueError("Method not recognized.")
    
    def create_metastatic_libraries_random(self, nsites, alter_prop, CN_coeff):
        init_drivers = [(chrname, r) for chrname, regions in self.drivers.items() for r in regions]

        ndrivers = len(init_drivers)
        met_libraries = [self]
        for _ in range(1, nsites):
            new_lib = copy.deepcopy(self)
            n_alt_drivers = min(np.random.binomial(ndrivers, alter_prop), ndrivers)
            idxs = np.random.choice(range(ndrivers), n_alt_drivers, replace=False)
            alt_drivers = [init_drivers[i] for i in idxs]
            for chrname, r in alt_drivers:
                new_lib.delta[chrname][r] = np.random.uniform(-CN_coeff, CN_coeff)
            new_lib.init_base_fit()
            new_lib.init_max_fit()
            met_libraries.append(new_lib)

        return met_libraries
    
    def create_metastatic_libraries_random_alt_drivers(self, nsites, alter_prop, CN_coeff, OG_r, TSG_r):
        """
        Similar to create_metastatic_libraries_random, creates metastatic site libraries by randomly altering a N driver regions, where N is drawn from a binomial with a mean of alter_prop. However, instead of resampling the coefficients of the chosen N drivers, it sets the coefficients to 0 and selects N random regions from among all possible.
        """
        OG_p = OG_r / (OG_r + TSG_r)

        prim_drivers = [(chrname, r) for chrname, regions in self.drivers.items() for r in regions]
        prim_neutrals = [(chrname, r) for chrname, num_regions in self.regions.items() for r in range(num_regions) if r not in self.drivers[chrname]]

        num_drivers = len(prim_drivers)
        num_neutrals = len(prim_neutrals)
        met_libraries = [self]
        for _ in range(1, nsites):
            new_lib = copy.deepcopy(self)
            num_former_drivers = np.random.binomial(int(num_drivers*alter_prop), 0.5)
            num_new_drivers = num_drivers - num_former_drivers

            former_driver_idxs = np.random.choice(range(num_drivers), num_former_drivers, replace=False)
            new_driver_idxs = np.random.choice(range(num_neutrals), num_new_drivers, replace=False)

            for i in former_driver_idxs:
                chrname, r = prim_drivers[i]
                new_lib.delta[chrname][r] = 0
                new_lib.drivers[chrname].remove(r)
            
            for i in new_driver_idxs:
                chrname, r = prim_neutrals[i]
                new_lib.drivers[chrname].append(r)
                if np.random.binomial(1, OG_p) == 1:
                    new_lib.delta[chrname][r] = np.random.uniform(0, CN_coeff)
                else:
                    new_lib.delta[chrname][r] = np.random.uniform(-CN_coeff, 0)

            for chrname in new_lib.drivers:
                new_lib.drivers[chrname].sort()

            new_lib.init_base_fit()
            new_lib.init_max_fit()
            met_libraries.append(new_lib)

        return met_libraries
    
    def create_metastatic_libraries_distance(self, nsites, alter_prop, CN_coeff, dists):
        init_drivers = [(chrname, a) for chrname in self.regions for a in [0, 1]]

        ndrivers = len(init_drivers)
        n_alt_drivers = np.random.binomial(ndrivers, alter_prop)
        dists = np.rint(dists*n_alt_drivers).astype(int)
        idxs = np.triu_indices(nsites)
        dmatrix = np.zeros((nsites, nsites), dtype=int)
        dmatrix[idxs] = dists[:, 0]
        dmatrix = dmatrix + dmatrix.T
        A, score = assign_n_alt_drivers(dmatrix, ndrivers)

        libraries = [self]
        unique = list(set.union(*[x for x in A.values()]))
        idxs = np.random.choice(range(ndrivers), len(unique), replace=False)
        alt_drivers = [init_drivers[i] for i in idxs]
        new_drivers = {}
        for i,(chrname, a) in zip(unique, alt_drivers):
            new_delta = np.random.uniform(-CN_coeff, CN_coeff)
            new_drivers[i] = (chrname, a, new_delta)

        for s in range(1, nsites):
            new_lib = copy.deepcopy(self)
            for i in A[s]:
                chrname, a, delta = new_drivers[i]
                new_lib.delta[chrname][a] = delta
            new_lib.init_base_fit()
            new_lib.init_max_fit()
            libraries.append(new_lib)

        return libraries
           
class RandomRegionLibrary(BaseRegionLibrary):
    """The region selection model with randomly generated selection coefficients.
    
    This model is similar to the chromosome-arm model but operates at the region/gene level. In particular, each reference region i on chromosome k is assigned a selection coefficient in range (-CN_coeff, CN_coeff), where positive values correspond to OGs, negative values correspond to tumor suppressor genes TSGs, and zero corresponds to neutral regions (NEU). The fitness :math:`s_a(c)` of cell :math:`c` in site :math:`a` is computed as

    .. math::
    
        s_a(c) = \\prod_{k=1}^K \\prod_{i=1}^{m_k} (1 + \\delta_{k,i})^{x_{k,i} / p_c},

    where :math:`K` is the number of chromosomes, :math:`m_k` is the number of reference regions on chromosome :math:`k`, the :math:`\\delta`'s are selection coefficients, :math:`x_{k,i}` is the copy number of region :math:`i` on chromosome :math:`k` in cell  :math:`c`, and :math:`p_c` is cell ploidy.

    """
    def initialize(
        self, 
        params: Optional[Parameters] = None, 
        CN_coeff: Optional[float] = None,
        OG_r: Optional[float] = None, 
        TSG_r: Optional[float] = None, 
        EG_r: Optional[float] = None
    ):
        """Method used to initialize selection coefficients. See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            params (Parameters, optional):
            CN_coeff (float, optional):
            OG_r (float, optional):
            TSG_r (float, optional):
            EG_r (float, optional):
        """
        params = fill_params(params, CN_coeff=CN_coeff, OG_r=OG_r, TSG_r=TSG_r, EG_r=EG_r)

        NEU_r = 1 - params.OG_r - params.TSG_r - params.EG_r
        for chrname, num_regions in self.regions.items():
            labels = np.random.choice([0, 1, 2, 3], size=num_regions, p=[params.OG_r, params.TSG_r, params.EG_r, NEU_r])
            for r,l in enumerate(labels):
                if l == 0:
                    self.delta[chrname].append(np.random.uniform(0, params.CN_coeff))
                    self.drivers[chrname].append(r)
                elif l == 1:
                    self.delta[chrname].append(np.random.uniform(-params.CN_coeff, 0))
                    self.drivers[chrname].append(r)
                elif l == 2:
                    self.delta[chrname].append(0)
                    self.essential[chrname].append(r)
                else:
                    self.delta[chrname].append(0)
        self.max_distinct_driv = int(self.max_distinct_driv_ratio * sum([len(self.drivers[chrname]) for chrname in self.drivers]))
        self.init_base_fit()
        self.init_max_fit()
    
class FittedRegionLibrary(BaseRegionLibrary):
    """Uses the same model as :code:`RandomRegionLibrary`, but the selection coefficients are given by the user.

    """
    def initialize(
        self, 
        filepath: Optional[str] = None,
        delta: Optional[Dict] = None,
    ):
        """Method used to initialize selection coefficients.

        Args:
            filepath (str, optional): Path to a tsv file containing the selection coefficients. The first entry is the chromosome name, followed by the selection coefficients in order of reference regions.
            delta (dict, optional): Instead of passing a file, users can pass a dictionary where keys are chromosome names and values are an :math:`m_k`-length array of coefficients.
        
        """
        if delta is not None:
            if len(set(delta.keys()) - set(self.chrom_lens.keys())) > 0:
                raise ValueError("Provided selection coefficient dictionary contains unrecognized chromosomes.")
            
            deltas = {}
            for chrname,v in delta.items():
                if not isinstance(v, (list, tuple, np.ndarray)):
                    raise ValueError("Provided selection coefficient dictionary values must be an array or list containing float values.")
                if len(v) > self.regions[chrname]:
                    raise ValueError(f"Number of deltas for {chrname} exceeds the number of regions specified by chrom_lens and region_len on library initialization.")
                for x in v:
                    if not isinstance(x, (float, np.floating)):
                        raise ValueError(f"Provided deltas are neither floats nor np.floats.")
                v = [float(x) for x in v]
                while len(v) != self.regions[chrname]:
                    v.append(0)
                deltas[chrname] = v
        
        elif filepath is not None:
            assert os.path.isfile(filepath), 'File does not exist at specified path.'
            deltas = {}
            with open(filepath) as f:
                for line in f:
                    info = line.strip().split('\t')
                    chrname = info[0]
                    if chrname not in self.chrom_lens:
                        raise ValueError("Provided arm selection coefficients file contains unrecognized chromosome.")
                    deltas[chrname] = [float(x) for x in info[1:]]

        else:
            raise ValueError("Must pass either filepath or delta.")
        
        for chrname in self.regions:
            if chrname not in deltas:
                deltas[chrname] = [0 for _ in range(self.regions[chrname])]
        self.delta = deltas
        
        for chrname,v in deltas.items():
            for i,x in enumerate(v):
                if x != 0:
                    self.drivers[chrname].append(i)

        self.max_distinct_driv = int(self.max_distinct_driv_ratio * sum([len(self.drivers[chrname]) for chrname in self.drivers]))
        self.init_base_fit()
        self.init_max_fit()


import numpy as np
from scipy.stats import expon
from itertools import chain
from collections import defaultdict
import logging
import pickle
import copy
import os

from typing import Optional, Union

from sistem.utilities.IO import setup_logger
from sistem.genome.genome import init_diploid_genome
from sistem.lineage import Cell, Clone
import sistem.lineage.mutation as mut
from sistem.lineage.lineage import create_clone_tree, create_converted_tree, create_singlecell_tree, resolve_multifurcation, merge_and_resolve_clone_tree, collapse_tree
from sistem.lineage.reevolve import evolve_tree_with_passengers
from sistem.lineage.utils import add_dummy_diploids
from sistem.anatomy.utils import get_num_migrations, draw_migrations, generate_migration_graph, save_migration_graph
from sistem.anatomy import BaseAnatomy
from sistem.data.profiles import save_mutations_from_tree, save_CNPs, save_SNVs, site_CN_averages
from sistem.parameters import Parameters, fill_params

class GrowthSimulator:
    """The main class used to run the simulator. 

    Args:
        anatomy (Anatomy): The Anatomy object which has been initialized. Required.

    Attributes:
        clones (Dict[list]): A dictionary where keys are site numbers (0-nsites-1), and values are a list of clones currently present in the site. 
        site_counts (array): Contains the total cell count of each site.
        gen (int): The current generation of the simulator.
    """
    def __init__(self, anatomy):
        """Constructor method

        """
        self.anatomy = self._check_anatomy(anatomy)

        self.clones = {i: [] for i in range(self.anatomy.nsites)}
        self.site_counts = np.zeros(self.anatomy.nsites, dtype=int)
        self.gen = 0

        self._empty_sites = set()
        self._clone_ids = [0 for _ in range(self.anatomy.nsites)]
        self._igenome = init_diploid_genome(self.anatomy.libraries[0])
        self._init_clone = Clone(name=self._assign_clone_name(0), genome=copy.deepcopy(self._igenome), library=self.anatomy.libraries[0], popsize=self.anatomy.N0, site=0, birth_gen=0)
        self.clones[0].append(self._init_clone)
        self.site_counts[0] += self._init_clone.popsize

        self._lifespan_gens = {i: defaultdict(list) for i in range(self.anatomy.nsites)}
        self._lifespan_gens[0][1].append(self._init_clone)

        # Data structures/variables used in sampling and phylogeny construction
        self._observed = defaultdict(int)
        self._clone_tree = None

    def __iter__(self):
        return chain.from_iterable(self.clones.values())

    def _check_anatomy(self, anatomy):
        if not isinstance(anatomy, BaseAnatomy):
            raise TypeError("Argument `anatomy' must be an instance of an Anatomy class.")
        if anatomy.initialized is False:
            raise ValueError("Anatomy object has not been properly initialized. Try running anatomy.initialize_distances method.")
        return anatomy
    
    def _assign_clone_name(self, site):
        name = self.anatomy.site_ids[site] + str(self._clone_ids[site])
        self._clone_ids[site] += 1
        return name
    
    def _get_mean_fitness(self):
        for clone in self:
            clone.update_fitness()
        mean_fits = []
        for s,clones in self.clones.items():
            if len(clones) == 0:
                mean_fits.append(0)
            else:
                mean_fits.append(np.average([clone.fitness for clone in clones], weights=[clone.popsize for clone in clones]))
        return mean_fits
    
    def _create_new_clones(self, parent_clone, site, distr_events, focal_rate, arm_rate, chromosomal_rate, WGD_rate, focal_gain_rate, chrom_dup_rate, length_mean, mag_mean, lifespan_mean):
        """
        Given array distr_events from draw_num_CNA_events, creates a new clone for each element with event count equal to the element
        """
        nviolations = 0
        for x,y in distr_events:
            clone = parent_clone.replicate(name=self._assign_clone_name(site))
            if x > 0:
                mut.select_CNA_events(clone, x, focal_rate, arm_rate, chromosomal_rate, WGD_rate, focal_gain_rate, chrom_dup_rate, length_mean, mag_mean)
            if y > 0:
                mut.select_SNV_events(clone, y)
            if clone.is_viable():
                self.clones[site].append(clone)
                if lifespan_mean > 1:
                    lifespan = max(1, round(expon.rvs(scale=2)))
                else:
                    lifespan = 1
                self._lifespan_gens[site][self.gen + lifespan].append(clone)
            else:
                nviolations += 1
                clone.die()
        return nviolations
    
    # PARALLELIZE?
    def _cycle_birthdeath(self, focal_driver_rate, arm_rate, chromosomal_rate, WGD_rate, focal_gain_rate, chrom_dup_rate, length_mean, mag_mean, SNV_driver_rate, lifespan_mean):
        totbirths, totdeaths = [0 for i in range(self.anatomy.nsites)], [0 for i in range(self.anatomy.nsites)]
        mean_fits = self._get_mean_fitness()
        
        # Driver mutations
        for s,m in enumerate(mean_fits):
            #clones = self.clones[s][:]
            clones = self._lifespan_gens[s][self.gen]
            if len(clones) > 1:
                print(self.gen, clones)
            if len(clones) > 0:
                Ntot, Etot = self.site_counts[s], self.anatomy.exp_pop[s][self.gen]
                #for clone in clones:
                while len(clones) > 0:
                    clone = clones.pop(0)
                    nbirths = clone.get_num_births(m, Ntot, Etot)
                    clone.birth_counts[self.gen] = nbirths
                    ndeaths = clone.popsize - nbirths
                    if nbirths > 0:
                        clone.popsize = 2*nbirths
                        clone_focal_rate = focal_driver_rate*clone.get_ploidy()/2
                        nevents_CN = mut.draw_num_CNA_events(nbirths, clone_focal_rate, arm_rate, chromosomal_rate, WGD_rate)
                        nevents_SNV = mut.draw_num_SNV_events(nbirths, SNV_driver_rate)
                        if nevents_CN > 0 or nevents_SNV > 0:
                            distr_events = mut.distribute_events(nbirths, nevents_CN, nevents_SNV)
                            nviolations = self._create_new_clones(clone, s, distr_events, clone_focal_rate, arm_rate, chromosomal_rate, WGD_rate, focal_gain_rate, chrom_dup_rate, length_mean, mag_mean, lifespan_mean)
                            clone.popsize -= len(distr_events)
                            self.site_counts[s] -= nviolations

                        if clone.popsize <= 0:
                            clone.die()
                            self.clones[s].remove(clone)
                        else:
                            if lifespan_mean > 1:
                                lifespan = max(1, round(expon.rvs(scale=2)))
                            else:
                                lifespan = 1
                            self._lifespan_gens[s][self.gen + lifespan].append(clone)
                    else:
                        clone.die()
                        self.clones[s].remove(clone)
                    self.site_counts[s] += nbirths - ndeaths
                    totbirths[s] += nbirths
                    totdeaths[s] += ndeaths
        
        # Passenger mutations
        #for clone in self:
        #    add_passengers(clone, focal_pass_rate, SNV_pass_rate)

        return mean_fits, totbirths, totdeaths

    def _cycle_migrations(self, t_max, pattern):
        # First figure out what migrations are happening, then afterwards do the shuffling
        totmigrations = [0 for i in range(self.anatomy.nsites)]
        if self.anatomy.nsites == 1:
            return totmigrations

        multi_source = True
        if 'S' in pattern:
            multi_source = False
        
        moves = defaultdict(list)
        for s,clones in self.clones.items():
            if len(clones) == 0:
                self._empty_sites.add(s)
            for clone in clones[:]:
                nmigrations = get_num_migrations(clone, self.anatomy.get_total_transition_probability(clone))
                if nmigrations > 0:
                    migrations = draw_migrations(nmigrations, self.anatomy.get_oneway_transition_probabilities(clone))
                    for a,occurs in enumerate(migrations):
                        if occurs == 1:
                            migration_size = 1
                            if multi_source:
                                migration_size = self.anatomy.N0

                            new_clone = clone.replicate(name=self._assign_clone_name(a), site=a, popsize=migration_size, library=self.anatomy.libraries[a])
                            moves[a].append(new_clone)
                            self.site_counts[a] += migration_size
                            self.site_counts[s] -= migration_size
                            clone.popsize -= migration_size

                    if clone.popsize <= 0:
                        self.clones[s].remove(clone)
                    totmigrations[s] += sum(migrations)
        
        for s,new_clones in moves.items():
            # For newly seeded sites, need to initialize growth functions
            if s in self._empty_sites:
                fits = []
                for clone in new_clones:
                    clone.update_fitness()
                    fits.append(clone.fitness)
                mean_fit = np.mean(fits)
                self.anatomy.init_pop_dynamic(s, self.gen, t_max, cur_fit=mean_fit, library=self.anatomy.libraries[s])
                self._empty_sites.remove(s)
            self.clones[s].extend(new_clones)
        
        return totmigrations

    # TODO: implement min detectable
    def _terminate(self, t_max, min_detectable):
        if sum(self.site_counts) == 0:
            logging.info('\nTERMINATE: No remaining living cells')
            return True
        elif self.gen > t_max:
            logging.info('\nTERMINATE: Max gen reached')
            return True
        elif np.all(self.site_counts >= min_detectable):
            logging.info('\nTERMINATE: Cell populations reached detectabled sizes in all sites')
            return True
        else:
            return False
    
    def _log_current_gen(self, mean_fits, totbirths, totdeaths, totmigrations):
        logging.info(f'\nGen{self.gen}')
        logging.info(f'P\t{len(self.clones[0])}\t{self.site_counts[0]}\t{mean_fits[0]}\t{totbirths[0]}\t{totdeaths[0]}\t{totmigrations[0]}')
        for s in range(1, self.anatomy.nsites):
            logging.info(f'M{s}\t{len(self.clones[s])}\t{self.site_counts[s]}\t{mean_fits[s]}\t{totbirths[s]}\t{totdeaths[s]}\t{totmigrations[s]}')

    def simulate_agents(
        self, 
        params: Optional[Parameters] = None,
        t_max: Optional[int] = None, 
        lifespan_mean: Optional[Union[int, float]] = None,
        min_detectable: Optional[int] = None, 
        focal_driver_rate: Optional[float] = None, 
        arm_rate: Optional[float] = None, 
        chromosomal_rate: Optional[float] = None, 
        WGD_rate: Optional[float] = None, 
        focal_gain_rate: Optional[float] = None, 
        chrom_dup_rate: Optional[float] = None, 
        length_mean: Optional[Union[int, float]] = None, 
        mag_mean: Optional[Union[int, float]] = None, 
        SNV_driver_rate: Optional[float] = None, 
        log_path: Optional[str] = None,
        pattern: str = ''
    ):
        """Begins the agent-based phase of the simulator. Runs until a minimum detectable number of cells is reached in each site, the max generation is reached, or no cells are left in any site. Only simulates driver mutations. See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            params (Parameters, optional):
            t_max (int, optional):
            lifespan_mean (int, float, optional):
            min_detectable (int, optional):
            focal_driver_rate (float, optional):
            arm_rate (float, optional):
            chromosomal_rate (float, optional):
            WGD_rate (float, optional):
            focal_gain_rate (float, optional):
            chrom_dup_rate (float, optional):
            length_mean (int, float, optional):
            mag_mean (int, float, optional):
            SNV_driver_rate (float, optional):
            log_path (str, optional): 
        """
        params = fill_params(params, t_max=t_max, lifespan_mean=lifespan_mean, min_detectable=min_detectable, focal_driver_rate=focal_driver_rate, arm_rate=arm_rate, chromosomal_rate=chromosomal_rate, WGD_rate=WGD_rate, focal_gain_rate=focal_gain_rate, chrom_dup_rate=chrom_dup_rate, length_mean=length_mean, mag_mean=mag_mean, SNV_driver_rate=SNV_driver_rate, log_path=log_path)
        setup_logger(file_path=params.log_path)

        if not self._init_clone.library.is_driver_SNV_model:
            params.SNV_driver_rate = 0

        if self.gen == 0:
            logging.info(f'STARTING AGENT GROWTH MODEL')
            logging.info('Site\tNumClones\tPopSize\tMeanFit\tNumBirths\tNumDeaths\tNumMigrations')
            mean_fits = totbirths = totdeaths = totmigrations = [0 for i in range(self.anatomy.nsites)]
            self._log_current_gen(mean_fits, totbirths, totdeaths, totmigrations)
        else:
            logging.info(f'\nCONTINUE')

        TERMINATE = self._terminate(params.t_max, params.min_detectable)
        while not TERMINATE:
            self.gen += 1
            Cell.gen = self.gen
            if self._terminate(params.t_max, params.min_detectable):
                self.gen -= 1
                Cell.gen = self.gen
                TERMINATE = True
                break
            mean_fits, totbirths, totdeaths = self._cycle_birthdeath(params.focal_driver_rate, params.arm_rate, params.chromosomal_rate, params.WGD_rate, params.focal_gain_rate, params.chrom_dup_rate, params.length_mean, params.mag_mean, params.SNV_driver_rate, params.lifespan_mean)
            totmigrations = self._cycle_migrations(params.t_max, pattern=pattern)
            self._log_current_gen(mean_fits, totbirths, totdeaths, totmigrations)
    
    def cycle_gen(
        self, 
        params: Optional[Parameters] = None,
        t_max: Optional[int] = None, 
        lifespan_mean: Optional[Union[int, float]] = None,
        min_detectable: Optional[int] = None, 
        focal_driver_rate: Optional[float] = None, 
        arm_rate: Optional[float] = None, 
        chromosomal_rate: Optional[float] = None, 
        WGD_rate: Optional[float] = None, 
        focal_gain_rate: Optional[float] = None, 
        chrom_dup_rate: Optional[float] = None, 
        length_mean: Optional[Union[int, float]] = None, 
        mag_mean: Optional[Union[int, float]] = None, 
        SNV_driver_rate: Optional[float] = None,
        log_path: Optional[str] = None,
        pattern: str = ''
    ):
        """Identical to the :code:`simulate_agents` method, but runs for a single generation only

        """
        params = fill_params(params, t_max=t_max, lifespan_mean=lifespan_mean, min_detectable=min_detectable, focal_driver_rate=focal_driver_rate, arm_rate=arm_rate, chromosomal_rate=chromosomal_rate, WGD_rate=WGD_rate, focal_gain_rate=focal_gain_rate, chrom_dup_rate=chrom_dup_rate, length_mean=length_mean, mag_mean=mag_mean, SNV_driver_rate=SNV_driver_rate, log_path=log_path)
        setup_logger(file_path=params.log_path)

        if self.gen == 0:
            logging.info(f'STARTING AGENT GROWTH MODEL')
            logging.info('Site\tNumClones\tPopSize\tMeanFit\tNumBirths\tNumDeaths\tNumMigrations')
            mean_fits = totbirths = totdeaths = totmigrations = [0 for i in range(self.anatomy.nsites)]
            self._log_current_gen(mean_fits, totbirths, totdeaths, totmigrations)
        else:
            logging.info(f'\nCONTINUE')
        
        if not self._terminate(params.t_max, params.min_detectable):
            self.gen += 1
            Cell.gen = self.gen
            mean_fits, totbirths, totdeaths = self._cycle_birthdeath(params.focal_driver_rate, params.arm_rate, params.chromosomal_rate, params.WGD_rate, params.focal_gain_rate, params.chrom_dup_rate, params.length_mean, params.mag_mean, params.SNV_driver_rate, params.lifespan_mean)
            totmigrations = self._cycle_migrations(params.t_max, pattern=pattern)
            self._log_current_gen(mean_fits, totbirths, totdeaths, totmigrations)

    def sample_cells(
        self, 
        params: Optional[Parameters] = None,
        ncells_prim: Optional[int] = None,
        ncells_meta: Optional[int] = None, 
    ):
        """Samples cells from each anatomical site. The probability that a cell is sampled from a clone is proportional to that clones population size. See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            params (Parameters, optional): 
            ncells_prim (int, optional): 
            ncells_meta (int, optional): 
        """
        params = fill_params(params, ncells_prim=ncells_prim, ncells_meta=ncells_meta)
        obs = [clone for clone in self._observed.keys()]
        for clone in obs:
            del self._observed[clone]
        
        for s in range(self.anatomy.nsites):
            if s == 0:
                ncells = params.ncells_prim
            else:
                ncells = params.ncells_meta
            clones = self.clones[s]
            pops = np.array([clone.popsize for clone in clones])
            tot_cells = sum(pops)
            if ncells > 0 and tot_cells >= ncells:
                idxs = np.random.choice(range(len(clones)), ncells, p=pops/tot_cells)
                for i in idxs:
                    self._observed[clones[i]] += 1
        self._clone_tree = create_clone_tree(self._observed)

    # Master function for creating resolved tree from sampled cells
    def simulate_singlecell_lineage(
        self, 
        params: Optional[Parameters] = None,
        out_dir: Optional[str] = None, 
        focal_pass_rate: Optional[float] = None, 
        SNV_pass_rate: Optional[float] = None, 
        focal_gain_rate: Optional[float] = None, 
        length_mean: Optional[Union[int, float]] = None, 
        mag_mean: Optional[Union[int, float]] = None, 
        bin_size: Optional[int] = None,
        ncells_normal: Optional[int] = None,
        ref: Optional[str] = None,
        alt_ref: Optional[str] = None
    ):
        """Construct a single-cell lineage from the sampled cells under a coalescent, constrained to obey the underlying clonal evolutionary history, and add passenger mutations. Requires running the :code:`sample_cells` class method beforehand. Automatically saves the ground truth tree, migration graph, mutation history, CNA profiles, and SNV profiles. See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            params (Params, optional): 
            out_dir (str, optional): 
            focal_pass_rate (float, optional): 
            SNV_pass_rate (float, optional): 
            focal_gain_rate (float, optional): 
            length_mean (int, float, optional): 
            mag_mean (int, float, optional): 
            bin_size (int, optional): 
            ncells_normal (int, optional): 
            ref (str, optional): 
            alt_ref (str, optional): 

        Returns:
            Tree: The single-cell lineage tree.
        """
        params = fill_params(params, out_dir=out_dir, focal_pass_rate=focal_pass_rate, SNV_pass_rate=SNV_pass_rate, focal_gain_rate=focal_gain_rate, length_mean=length_mean, mag_mean=mag_mean, bin_size=bin_size, ncells_normal=ncells_normal, ref=ref, alt_ref=alt_ref)

        if len(self._observed) == 0:
            raise ValueError("Must run GrowthSimulator.sample_cells beforehand.")

        tree = create_singlecell_tree(self._clone_tree, self._observed, self.gen)
        resolve_multifurcation(tree)
        evolve_tree_with_passengers(tree, self._igenome, params.focal_pass_rate, params.SNV_pass_rate, params.focal_gain_rate, params.length_mean, params.mag_mean)
        if params.ncells_normal >= 1:
            add_dummy_diploids(tree, self.gen, params.ncells_normal)
        tree.save(os.path.join(params.out_dir, 'cell_tree_full.nwk'))
        collapse_tree(tree)
        tree.save(os.path.join(params.out_dir, 'cell_tree.nwk'))
        if self.anatomy.nsites > 1:
            edges = generate_migration_graph(tree, self.anatomy.site_ids)
            save_migration_graph(params.out_dir, edges)
        save_mutations_from_tree(tree, params.out_dir, params.bin_size)
        save_CNPs(tree, params.out_dir, params.bin_size)
        save_SNVs(tree, params.out_dir)
        return tree

    def simulate_clonal_lineage(
        self, 
        params: Optional[Parameters] = None,
        out_dir: Optional[str] = None, 
        min_mut_fraction: Optional[float] = None, 
        focal_pass_rate: Optional[float] = None, 
        SNV_pass_rate: Optional[float] = None, 
        focal_gain_rate: Optional[float] = None, 
        length_mean: Optional[Union[int, float]] = None, 
        mag_mean: Optional[Union[int, float]] = None, 
        bin_size: Optional[int] = None,
        ncells_normal: Optional[int] = None,
        ref: Optional[str] = None,
        alt_ref: Optional[str] = None
    ):
        """Construct a clonal lineage from the sampled cells and adds passenger mutations. Optionally prune the observed clones based on a minimum genotype frequency threshold. Requires running the :code:`sample_cells` class method beforehand. Automatically saves the ground truth tree, migration graph, mutation history, CNA profiles, and SNV profiles. See :ref:`Parameters <parameters>` for an explanation of the parameters.

        Args:
            params (Params, optional): 
            out_dir (str, optional): 
            min_mut_fraction (float, optional): 
            focal_pass_rate (float, optional): 
            SNV_pass_rate (float, optional): 
            focal_gain_rate (float, optional): 
            length_mean (int, float, optional): 
            mag_mean (int, float, optional): 
            bin_size (int, optional): 
            ncells_normal (int, optional): 
            ref (str, optional): 
            alt_ref (str, optional): 

        Returns:
            Tree: The clonal lineage tree.
        """
        params = fill_params(params, out_dir=out_dir, min_mut_fraction=min_mut_fraction, focal_pass_rate=focal_pass_rate, SNV_pass_rate=SNV_pass_rate, focal_gain_rate=focal_gain_rate, length_mean=length_mean, mag_mean=mag_mean, bin_size=bin_size, ncells_normal=ncells_normal, ref=ref, alt_ref=alt_ref)

        if len(self._observed) == 0:
            raise ValueError("Must run GrowthSimulator.sample_cells beforehand.")

        tree, obs = create_converted_tree(self._clone_tree, self._observed)
        sample_counts = merge_and_resolve_clone_tree(tree, obs, self.anatomy.nsites, self.gen, params.min_mut_fraction)
        evolve_tree_with_passengers(tree, self._igenome, params.focal_pass_rate, params.SNV_pass_rate, params.focal_gain_rate, params.length_mean, params.mag_mean)
        if params.ncells_normal >= 1:
            add_dummy_diploids(tree, self.gen, 1)
            dip = tree.find('diploid')
            assert dip != None, "Cannot find diploid node..."
            dip.info['count'] = params.ncells_normal
        tree.save(os.path.join(params.out_dir, 'clone_tree_full.nwk'))
        collapse_tree(tree)
        tree.save(os.path.join(params.out_dir, 'clone_tree.nwk'))
        for cell,count in sample_counts.items():
            cell.info['count'] = count
        if self.anatomy.nsites > 1:
            edges = generate_migration_graph(tree, self.anatomy.site_ids)
            save_migration_graph(params.out_dir, edges)
        save_mutations_from_tree(tree, params.out_dir, params.bin_size)
        save_CNPs(tree, params.out_dir, params.bin_size)
        save_SNVs(tree, params.out_dir, ref=params.ref, alt_ref=params.alt_ref)
        site_CN_averages(tree, params.out_dir, params.bin_size)
        return tree
    
    def save_checkpoint(self, outfile):
        """Saves the growth simulator in its current state to a file.

        Args:
            outfile (str): The location to save the growth simulator.
        """
        with open(outfile, 'wb') as f:
            pickle.dump(self, f)
    

